import django.forms
import datetime
import decimal

from collections import OrderedDict
from typing import List, Type, Optional
from django.db.models import Model, ForeignKey, UUIDField, ManyToManyField
from django.db.models.fields.files import FieldFile
from django.db.models.query import ValuesListIterable, QuerySet
from django.apps import apps

from . import LoggerException
from . import models


def get_models() -> List[Type[Model]]:
    """
    :return: models under logging (have LOG_NAME attribute)
    """
    models = []

    for model in apps.get_models():
        if hasattr(model, 'LOG_NAME') and model.LOG_NAME:
            models.append(model)

    return models


def get_model_by_log_name(log_name) -> Type[Model]:
    for item in get_models():
        if item.LOG_NAME == log_name:
            return item
    raise LoggerException('Model with LOG_NAME "%s" not found' % log_name)


def get_models_choices() -> List:
    choices = []
    for model in get_models():
        choices.append((model.LOG_NAME, model._meta.verbose_name))
    return choices


def object_from_log(log: models.Log) -> Optional[Model]:
    """
        create model object in memory from log data
    """
    object_data = log.get_object_data()

    if not object_data:
        return

    model = get_model_by_log_name(log.object_name)

    fields = {}
    for field in model._meta.fields:
        fields[field.name] = field

    model_object = model()

    # first - "id", for many-to-many recreation
    if 'id' in fields:
        model_object.id = object_data['id']

    # other fields
    for field_name, value in object_data.items():
        # skip other model fields (for example, ManyToMany)
        if field_name not in fields:
            continue

        # check is field FK
        if isinstance(fields.get(field_name), ForeignKey):
            # for FK fields add "_id" to field name
            field_name += '_id'

        setattr(model_object, field_name, value)

    return model_object


def object_display_from_log(log: models.Log) -> Optional[dict]:
    """
    :param log:
    :return: human oriented object data representation (use verbose names and etc)
    """
    if not log.object_data:
        return

    display_data = OrderedDict()
    object_data = log.get_object_data()
    model = get_model_by_log_name(log.object_name)

    for field in model._meta.fields:
        value = object_data.get(field.name)

        if isinstance(field, ForeignKey):
            filter_params = {}
            field_rel = field.rel if hasattr(field, 'rel') else field.remote_field
            filter_params[field_rel.field_name] = value
            if hasattr(field_rel, 'to'):
                related_qs = field_rel.to.objects.filter(**filter_params)
            else:
                related_qs = field_rel.field.related_model.objects.filter(**filter_params)
            if len(related_qs):
                value = related_qs[0]

        if isinstance(value, list):
            value = ', '.join(value)

        display_data[field.verbose_name] = value

    for field in model._meta.many_to_many:
        value = object_data.get(field.name, [])
        related_qs = field.related_model.objects.filter(pk__in=value)
        display_data[field.verbose_name] = ', '.join(sorted([str(i) for i in related_qs]))

    return display_data


def get_changes_display(previous_log: models.Log, current_log: models.Log) -> Optional[dict]:
    """
    :param previous_log:
    :param current_log:
    :return: human oriented representation of differences between objects data two of log records
    """
    previous_display = previous_log.get_object_data_display()
    current_display = current_log.get_object_data_display()

    if not previous_display or not current_display:
        return

    changes_display = OrderedDict()
    for field_name in current_display.keys():
        if current_display[field_name] != previous_display.get(field_name):
            changes_display[field_name] = current_display[field_name]

    return changes_display


def model_to_dict(obj: Model) -> dict:
    obj_dict = django.forms.model_to_dict(obj)
    obj_dict = _converter(obj_dict)

    # model_to_dict from django.forms skip some fields... Add it.
    for field in obj._meta.fields:
        if isinstance(field, UUIDField):
            obj_dict[field.name] = str(getattr(obj, field.name))

    # replace m2m object instance by pk
    for field_name, value in obj_dict.items():
        if isinstance(value, list):
            new_value = []
            for item in value:
                if isinstance(item, Model):
                    new_value.append(item.pk)
                else:
                    new_value.append(item)
            obj_dict[field_name] = new_value

    return obj_dict


def _converter(obj_dict: dict) -> dict:
    for key, value in obj_dict.items():
        if isinstance(value, dict):
            obj_dict[key] = _converter(value)

        elif isinstance(value, FieldFile):
            obj_dict[key] = str(value)

        elif isinstance(value, datetime.date):
            obj_dict[key] = value.isoformat()

        elif isinstance(value, datetime.time):
            obj_dict[key] = value.isoformat()

        elif isinstance(value, datetime.datetime):
            obj_dict[key] = value.isoformat(sep=' ')

        elif isinstance(value, ValuesListIterable):
            obj_dict[key] = list(value)

        elif isinstance(value, QuerySet):
            obj_dict[key] = list(value.values_list('pk', flat=True))

        elif isinstance(value, bytes):
            obj_dict[key] = value.decode('utf-8')

        elif isinstance(value, decimal.Decimal):
            obj_dict[key] = [str(i) for i in [value]]

    return obj_dict



import django_filters
from . import models


class Log(django_filters.FilterSet):
    datetime_from = django_filters.DateTimeFilter('created', lookup_expr='gte')
    datetime_to = django_filters.DateTimeFilter('created', lookup_expr='lte')
    message = django_filters.CharFilter('message', lookup_expr='icontains')
    action = django_filters.MultipleChoiceFilter('action', choices=models.Log.ACTION_CHOICES)
    level = django_filters.MultipleChoiceFilter('level', choices=models.Log.LOG_LEVEL_CHOICES)

    class Meta:
        model = models.Log
        fields = ['datetime_from', 'datetime_to', 'action', 'level', 'object_name', 'message', 'username',]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-08-10 11:27
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sw_logger', '0003_auto_20170624_1913'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='log',
            options={'ordering': ('id',)},
        ),
        migrations.AddField(
            model_name='log',
            name='fk_object_id',
            field=models.IntegerField(db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='log',
            name='action',
            field=models.CharField(choices=[('', ''), ('created', 'created'), ('updated', 'updated'), ('deleted', 'deleted'), ('other', 'other')], default='', max_length=10),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2017-06-24 16:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sw_logger', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='log',
            name='http_general',
        ),
        migrations.AddField(
            model_name='log',
            name='http_method',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='log',
            name='http_path',
            field=models.TextField(blank=True),
        ),
    ]



import json
import pprint
from typing import Optional
from django.db import models
from django.contrib.auth import get_user_model
from . import consts


class Log(models.Model):
    ACTION_CHOICES = [('', '')] + list(consts.ACTION_CHOICES)
    LOG_LEVEL_CHOICES = [('', '')] + list(consts.LOG_LEVEL_CHOICES)

    action = models.CharField(max_length=10, choices=ACTION_CHOICES, default='')
    message = models.CharField(max_length=255)
    func_name = models.CharField(max_length=255)
    level = models.CharField(max_length=10, choices=LOG_LEVEL_CHOICES, default='NOTSET')

    http_path = models.TextField(blank=True)
    http_method = models.TextField(blank=True)
    http_request_get = models.TextField(blank=True)
    http_request_post = models.TextField(blank=True)
    http_referrer = models.CharField(max_length=255, blank=True)

    user_id = models.IntegerField(db_index=True, null=True)
    username = models.CharField(max_length=255, db_index=True, blank=True)

    object_name = models.CharField(max_length=255, db_index=True, blank=True)
    object_id = models.IntegerField(db_index=True, null=True)
    fk_object_id = models.IntegerField(db_index=True, null=True)
    object_data = models.TextField(blank=True)

    extra = models.TextField(blank=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('id', )

    def get_model_object(self) -> models.Model:
        from . import tools

        if not hasattr(self, '_got_model_object'):
            self._got_model_object = tools.object_from_log(self)

        return self._got_model_object

    def get_object_model_name(self) -> Optional[str]:
        model_object = self.get_model_object()
        if not model_object:
            return

        name = model_object._meta.verbose_name
        return name

    def get_object_data(self) -> Optional[dict]:
        if not self.object_data:
            return

        return json.loads(self.object_data)

    def get_object_data_display(self) -> Optional[dict]:
        from . import tools
        return tools.object_display_from_log(self)

    def get_user(self):
        if not self.user_id:
            return

        User = get_user_model()
        return User.objects.get(id=self.user_id)

    def get_previous_object_log(self) -> Optional['Log']:
        """
        :return: previous log record for same object (same object_name and object_id)
        """
        if not self.object_data:
            return

        previous_qs = Log.objects.filter(
            id__lt=self.id,
            object_name=self.object_name,
            object_id=self.object_id,
        )
        if previous_qs:
            return previous_qs.last()

    def get_changes(self) -> Optional[dict]:
        from . import tools

        previous_object_log = self.get_previous_object_log()
        if not previous_object_log:
            return self.get_object_data_display()

        return tools.get_changes_display(previous_object_log, self)

    def get_extra_data(self):
        """
        try parse json from extra-field
        """
        if not self.extra:
            return

        try:
            return json.loads(self.extra)
        except Exception:
            pass

    def get_extra_pretty(self):
        """
        pretty output extra-field data
        """
        extra_data = self.get_extra_data()
        if not extra_data:
            return

        return pprint.pformat(extra_data)